"""------------------------------------------------------------------------------------------
Bayesian Affect Control Theory 
Agent class with emotions 
Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey
December 2014
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
Research sponsored by the Natural Sciences and Engineering Council of Canada (NSERC).
use python 2
see README for details
----------------------------------------------------------------------------------------------"""

import os
import math
import sys
import re
import copy
import numpy as NP
import itertools
import time
from pomcp import *
sys.path.append("./gui/")
from cEnum import eTurn

from bayesact import *

#the file with emotion dictionary in it
fmfname = "fmodifiers.dat"

emfname=os.path.join(os.path.dirname(__file__), "temotions-male.dat")
effname=os.path.join(os.path.dirname(__file__), "temotions-female.dat")


#open the input file for emotions
#which contains a binary number giving the terms in (fae,fap,faa,ea,ep,ea) where (ea,ep,ea) is the emotions vector
#followed by 3 coefficients
#the first row gives the constant d
#the next three rows give the 3x3 matrix E
#the next three rows give the 3x3 matrix R
#the final row gives the diagonal of the diagonal matrix Q
#for males
fb=file(emfname,"rU")
emot_d=map(lambda(z): float(z),fb.readline().split()[1:])
#print emot_d
emot_E=[]
for i in range(3):
    emot_E.append(map(lambda(z): float(z),fb.readline().split()[1:]))
#actually the transpose
emot_E = NP.array(emot_E).transpose()
#print emot_E
emot_R=[]
for i in range(3):
    emot_R.append(map(lambda(z): float(z),fb.readline().split()[1:]))
emot_R = NP.array(emot_R).transpose()
#print emot_R
#the rest is the Q matrix - 
#this is read generally as follows: 
#Z100XXX means Q_e and multiples fa_e
#Z100010 means the coefficient of e_e multiples fa_p
#Z010XXX means Q_p and multiplies fa_p
#...etc
#we will assume only one fa and one e multiplier per matrix row
emot_Q = {}
for l in ['e','p','a']:
    emot_Q[l] = []
    for i in range(3):
        emot_Q[l].append([0.0,0.0,0.0])
x = fb.readline()
while x!="":
    y=x.split()
    #the f value to use is a 1 in this vector
    fvec=list(y[0])[4:7]
    fvecc=[w[0] for w in filter(lambda (z): z[1]=="1", zip(fvars[0:3],fvec))]
    #fvecc has a string like Fae in it - the third letter gives us which Q matrix this corresponds to
    #the tau dimension we assign to  is given by the position of the 1 in this vector
    evec = list(y[0])[1:4]
    emot_Q[fvecc[0][2]][evec.index('1')] = map(lambda(z): float(z),y[1:])
    x = fb.readline()
fb.close()

for l in ['e','p','a']:
    emot_Q[l] = NP.array(emot_Q[l]).transpose()

#print emot_Q

def computeEmotion(fa,ta):
    normfact = copy.deepcopy(emot_E)
    #possibly don't do the ones that are all zeros here
    for (i,e) in zip([0,1,2],['e','p','a']):
        normfact += NP.diag([fa[i]]*3)*emot_Q[e]
    #print "emot_E is: "
    # print emot_E
    #print "normfact is: "
    #print normfact
    #print "ta-Rfa-d is: "
    #print ta-NP.dot(emot_R,fa)-emot_d
    return NP.dot(NP.linalg.inv(normfact),(ta-NP.dot(emot_R,fa)-emot_d))
        



#now, we should here remove any 'emot_Q' matrices that are all zeros and set them to null
#as we don't want to have to keep re-doing those computations every time

# An EmotionalAgent is an agent that has an additional 3D component for actions, eab, 
# and an additional observation of emotions, 
class EmotionalAgent(Agent):


    def __init__(self,*args,**kwargs):

        #std.dev in the output covariance for emotions
        self.gammae_value = kwargs.get("gammae_agent",0.1) 

        super(EmotionalAgent, self).__init__(self,*args,**kwargs)


        
        self.emotdict = readSentiments(fmfname,self.agent_gender)
        self.emotion = {}
        self.emotion["agent"]=NP.zeros(3)
        self.emotion["client"]=NP.zeros(3)


    

    #really the current emotion should be added into the State ... 
    def currentEmotion(self,state,who):
        if who=="agent":
            fa = state.f[0:3]
            ta = state.tau[0:3]
        else:
            fa = state.f[6:9]
            ta = state.tau[6:9]
        return computeEmotion(fa,ta)

    #who can be "agent" or "client"
    def expectedEmotion(self,who):
        self.emotion[who] = NP.zeros(3)
        sumw=0
        for sample in self.samples:
            theemot = self.currentEmotion(sample,who)
            self.emotion[who] = NP.add(self.emotion[who],sample.weight*theemot)
            sumw += sample.weight
        self.emotion[who] /= sumw
        return self.emotion[who]

    def findNearestEmotion(self,curremot):
        return findNearestEPAVector(curremot,self.emotdict)
    
    def lookUpEmotionLabel(self,emot_label):
        if self.emotdict.has_key(emot_label):
            return NP.asarray([float(self.emotdict[emot_label]["e"]),
                               float(self.emotdict[emot_label]["p"]),
                               float(self.emotdict[emot_label]["a"])])
        else:
            return []

    def init_output_covariances(self):

        super(EmotionalAgent, self).init_output_covariances()

        self.gammae_value2=self.gammae_value*self.gammae_value

        #precomputed stuff for computing normal pdfs for 3D vectors
        theivar=1.0/float(self.gammae_value2)
        thedet=math.pow(float(self.gammae_value),3)
        self.ldenom_e=math.log(math.pow((2*NP.pi),1.5)*thedet)
        self.ivar_e = theivar*0.5


    def evalSampleEvar(self,state,turn,observ):
        weight=0.0
        if (not observ==[]) and turn=="client":
            curr_emot = self.currentEmotion(state,"client")
            dvo=NP.array(curr_emot)-NP.array(observ)
            weight += self.normpdf(dvo,0.0,self.ivar_e,self.ldenom_e)
        else:
            weight=0.0
        return weight

    #the main SMC update step. Given an affective action aab and a propositional action paab,
    #and an affective observation observ and an  observation xobserv, update the samples
    #by first drawing an unweighted set, propagating those forward, computing the weights
    #and finally computing the deflection
    #I copied this over verbatim from bayesact.py ... really this should have been done better but
    #this function needed to be broken down more there so we could just overload the weighting part here
    #could (should) easily move this back to Agent class (bayesact.py) since eobserv is last argument default None
    #also put in a stub evalSampleEvar there that just returns 0.0 and override it here in EmotionalAgent - simple!
    def propagate_forward(self,aab,observ,xobserv=[],paab=None,verb=False,plotter=None,agent=eTurn.learner,eobserv=None):

        if False and verb:
            print 100*"!","samples before resample:"
            printSamples(self.samples)

        #resample to get an unweighted set
        new_samples=drawSamples(self.samples)

        #add "jitter" here instead of the proposal idea...
        #possibly just add the jitter to the client identity
        if self.agent_rough>0 or self.client_rough>0:
            self.roughenSamples(new_samples)

        #this is where the samples can be extracted for plotting
        if verb:
            print 100*"!","unweighted set:"
            printSamplesFW(new_samples)

        #send new samples to plotter
        if (None != plotter):
            self.sendSamplesToPlotter(new_samples,plotter,agent)
            
            if (eTurn.learner == agent):
                plotter.m_LearnerPreviousAction = aab
            else:
                plotter.m_SimulatorPreviousAction = aab

        #propagate forward
        self.samples=[]
        totweight=0.0
        for sample in new_samples:
            fVarSampler = FVarSampler(self.isiga, self.isigf, sample.tau, \
                                      sample.f, self, sample.get_turn(), \
                                      aab, observ)

            newsample = self.sampleNext(fVarSampler, fvars, tvars, sample, \
                                        aab, paab)

            theweight = self.evalSampleFvar(fvars,self.tdyn,newsample,self.theivar,self.ldenom,sample.get_turn(),observ)
            newweight = newsample.weight+theweight

            theweight = self.evalSampleEvar(newsample,sample.get_turn(),eobserv)
            newweight = newweight+theweight

            thexweight = self.evalSampleXvar(newsample,xobserv,sample)


            newsample.weight=math.exp(newweight)*thexweight
            totweight += newsample.weight
            self.samples.append(newsample)

                
        print 20*"%"," total weight: ",totweight
        number_of_reset_attempts=0
        while totweight<1e-20:
            #reset - the total weight has collapsed and the particle filter is broken
            print 100*"r","eset"
            #we used to do it this way
            #tmp_avgs=self.initialise_array(self.init_f,self.init_prop,self.init_initx)
            #now, draw a brand new set of samples, but using observ as if it was set exactly
            #can do this using turn="reset" (special case)
            self.samples=[]
            totweight=0.0
            for sample in new_samples:
                fVarSampler = FVarSampler(self.isiga, self.isigf, sample.tau, \
                                          sample.f, self, "creset", \
                                          aab, observ)

                newsample = self.sampleNext(fVarSampler, fvars, tvars, sample, \
                                            aab, paab)
                #here we don't compute the weight at all using evalSampleFvar since it may be totally wrong
                theweight = 0.0
                newweight = newsample.weight+theweight

                thexweight = self.evalSampleXvar(newsample,xobserv,sample)

                #now, sometimes this fails and returns zero if all the "turn" indicators in the previous
                #sample disagree with xobserv, and we are using an evaluation of the turn where the previous
                #sample is used as the X value for the weight.  In such cases, we will assume that at least
                #some of the new sample x values are "correct" and we'll set the weight to acccordingly
                #or we can just set all weights to one and hope for the best
                if thexweight < 1e-20:
                    thexweight = 1.0
                newsample.weight=math.exp(newweight)*thexweight
                totweight += newsample.weight
                self.samples.append(newsample)
                    
            number_of_reset_attempts += 1



        if False and verb:
            print 100*"!","new samples:"
            printHeavySamples(self.samples,0.001)

        avgstate=sampleMean(self.samples)
        self.deflection_avg=NP.dot(avgstate.tau-avgstate.f,avgstate.tau-avgstate.f)
        self.x_avg=avgstate.x


        #prune POMCP tree if we are doing pomcp
        if self.use_pomcp:
            #possibly allow the user the explore the POMCP tree
            if self.pomcp_doInteractive:
                askq=raw_input("POMCP about to prune  - do you want to interactively explore the tree? (enter means no, anything else is yes):")
                if not askq == "":
                    self.pomcp_agent.POMCP_interactiveExplorePlanTree()
            #prune the tree
            print 100*"~~"
            print aab,paab,observ,xobserv
            self.pomcp_pruneTree((aab,paab),(observ,xobserv))

        return avgstate



if __name__ == "__main__":
    

    learn_agent=EmotionalAgent()
    #here we are just going to use the emotion calculation functions
    fa=[ 2.16291926,  2.11364315, -0.42280131]
    ta=[.81019615,  0.47730238,  0.10942384] 

    fa = [1.47402725,  0.74727138,  0.53023657]
    ta = [0.84485561, -0.94907817,  0.90339819]


    fa = [1.52991219,  0.8353706,   0.52557234]
    ta = [2.20050925, -1.02906862,  0.93547473]

    fa = [1.60078992,  2.5765918,  -0.48768776] 
    ta = [1.4138979,   1.3309068,   0.28254774]

    fa = [1.9383111,   2.2320633,  -0.89562094]
    ta = [0.21257395,  1.48223805, -0.12987478]

    fa = [1.33251104,  1.40156676,  1.44295723]
    ta = [0.80573881, -0.66668066,  0.61625101]


    fa = [2.34713568,  1.43108062,  2.53255505]
    ta = [1.62898565, -0.17052302,  0.24915186]
    fa=[0.92770291,  1.23166074,  0.95659071]
    ta=[1.3731803,  -0.05628433,  0.26240033]
    curremot = computeEmotion(fa,ta)
    print learn_agent.findNearestEmotion(curremot)
    print curremot
    


